#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: validate

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    7/10/18
"""
import daiquiri

from metapype.eml import rule
from metapype.eml.exceptions import MetapypeRuleError, UnknownNodeError, ChildNotAllowedError
from metapype.eml.validation_errors import ValidationError
from metapype.model.node import Node


logger = daiquiri.getLogger("validate: " + __name__)


def node(n: Node, errs: list = None) -> None:
    """
    Validates a given node for rule compliance.

    Args:
        n: Node instance to be validated
        errs: List container for validation errors (fail fast if None)

    Returns:
        None

    Raises:
        MetapypeRuleError: An unknown type of node for EML
    """
    if n.name not in rule.node_mappings:
        msg = f"Unknown node rule type: {n.name}"
        if errs is None:
            raise UnknownNodeError(msg)
        else:
            errs.append((ValidationError.UNKNOWN_NODE, msg, n))
    else:
        node_rule = rule.get_rule(n.name)
        node_rule.validate_rule(n, errs)


def prune(n: Node, strict: bool = False) -> list:
    """
    Prune in place all non-valid nodes from the tree

    Args:
        n: Node
        strict:

    Returns: List of pruned nodes

    Side-effects: Non-valid nodes are pruned from the tree

    """
    pruned = list()
    if n.name != "metadata":
        try:
            node(n)
        except UnknownNodeError as ex:
            logger.debug(f"Pruning: {n.name}")
            pruned.append((n, str(ex)))
            if n.parent is not None:
                n.parent.remove_child(n)
            Node.delete_node_instance(n.id)
            return pruned
        except ChildNotAllowedError as ex:
            r = rule.get_rule(n.name)
            children = n.children.copy()
            for child in children:
                if not r.is_allowed_child(child.name):
                    logger.debug(f"Pruning: {child.name}")
                    pruned.append((child, str(ex)))
                    n.remove_child(child)
                    Node.delete_node_instance(child.id)
        except MetapypeRuleError as ex:
            logger.debug(ex)
        children = n.children.copy()
        for child in children:
            pruned += prune(child, strict)
            if strict and child not in pruned:
                try:
                    node(child)
                except MetapypeRuleError as ex:
                    logger.debug(f"Pruning: {child.name}")
                    pruned.append((child, str(ex)))
                    n.remove_child(child)
                    Node.delete_node_instance(child.id)
    return pruned


def tree(n: Node, errs: list = None) -> None:
    """
    Recursively walks from the root node and validates
    each child node for rule compliance.

    Args:
        n: Node instance of root for validates
        errs: List container for validation errors (fail fast if None)

    Returns:
        None
    """
    node(n, errs)
    if n.name != "metadata":
        for child in n.children:
            tree(child, errs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: node

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    5/15/18
"""
import copy
from enum import Enum
import json
import uuid

import daiquiri


logger = daiquiri.getLogger("node.py: " + __name__)


class Shift(Enum):
    LEFT = 0
    RIGHT = 1


class Node(object):

    store = {}

    def __init__(self, name: str, id: str = None, parent=None, content: str = None):
        """
        Model node class representation.

        Attributes:
            name: Required string of metadata element name being modeled
            id: Optional node identifier
            parent: Optional parent node
            content: Optional string content
        """

        self._id = str(uuid.uuid1()) if id is None else id
        self._name = name
        self._parent = parent
        self._content = None if content is None else str(content)
        self._tail = None
        self._attributes = {}
        self._nsmap = {}
        self._prefix = None
        self._extras = {}
        self._children = []
        Node.set_node_instance(self)

    def __str__(self):
        s = json.dumps(self.__object(), indent=2)
        return s

    def __repr__(self):
        r = self.__object()
        return str(r)

    def __object(self):
        o = {self._name: []}
        o[self._name].append({"id": self._id})
        o[self._name].append({"parent": self._parent.name if self._parent else None})
        o[self._name].append({"parent_id": self._parent.id if self._parent else None})
        o[self._name].append({"nsmap": self._nsmap})
        o[self._name].append({"prefix": self._prefix})
        o[self._name].append({"attributes": self._attributes})
        o[self._name].append({"extras": self._extras})
        o[self._name].append({"content": self._content})
        o[self._name].append({"tail": self._tail})
        children = []
        for child in self._children:
            children.append(child.name)
        o[self._name].append({"children": children})
        return o

    @classmethod
    def delete_node_instance(cls, id: str, children: bool = True):
        """
        Removes the node instance from the store; if children set to True, then
        remove all children recursively.

        Args:
            id: str node identifier
            children: bool

        Returns:
            None
        """
        if children:
            node = cls.get_node_instance(id)
            for child in node.children:
                cls.delete_node_instance(child.id)
        del Node.store[id]

    @classmethod
    def fix_nsmap(cls, node: "Node", nsmap: dict = None, nsmap_id: int = None) -> None:
        """
        Fixes the namespace mappings for the node and its children
        Args:
            node: Node - the node of which to fix the namespace map
            nsmap: Dict - the namespace map of the node's parent
            nsmap_id: Int - the object identifier of the node's nsmap; to be used for
                            repeated nsmap objects

        Returns:
            None
        """
        if nsmap is not None:
            if nsmap_id is None:
                nsmap_id = id(node.nsmap)
            if nsmap == node.nsmap:
                node.nsmap = nsmap
            else:
                for prefix in nsmap:
                    if prefix not in node.nsmap:
                        node.nsmap = copy.deepcopy(node.nsmap)
                    node.nsmap[prefix] = nsmap[prefix]

        for child in node.children:
            if id(child.nsmap) == nsmap_id:
                child.nsmap = node.nsmap
                cls.fix_nsmap(child, node.nsmap, nsmap_id=nsmap_id)
            else:
                cls.fix_nsmap(child, node.nsmap)

    @classmethod
    def get_node_instance(cls, id: str) -> "Node":
        """
        Returns the instance of a node from its identifier

        Args:
            id: Str

        Returns:
            Node
        """
        return cls.store.get(id, None)

    @classmethod
    def set_node_instance(cls, node: "Node"):
        """
        Sets the node instance in the node store

        Args:
            node:

        Returns:
            None
        """
        cls.store[node.id] = node

    def add_attribute(self, name, value):
        self._attributes[name] = value

    def add_child(self, child, index=None) -> None:
        """
        Adds a child object into the children list at either the end of
        the list or at the index location if specified, and assigns the
        parent namespace map to the child.
        Args:
            child: Node
            index: Int
        Returns:
            None
        """
        if index is None:
            self._children.append(child)
            child.parent = self
        else:
            self._children.insert(index, child)
            child.parent = self

        if self.nsmap == child.nsmap:
            child.nsmap = self.nsmap
        else:
            for prefix in self.nsmap:
                if prefix not in child.nsmap:
                    child.add_namespace(prefix, self.nsmap[prefix])

    def add_extras(self, key: str, value: str):
        self._extras[key] = value

    def add_namespace(self, prefix: str, namespace: str, nsmap_id: int = None):
        if nsmap_id is None:
            nsmap_id = id(self.nsmap)
        if prefix in self.nsmap:
            self.nsmap[prefix] = namespace
        else:
            self.nsmap = copy.deepcopy(self.nsmap)
            self.nsmap[prefix] = namespace

        for child in self._children:
            if id(child.nsmap) == nsmap_id:
                child.nsmap = self.nsmap
                child.add_namespace(prefix, namespace, nsmap_id=nsmap_id)
            else:
                child.add_namespace(prefix, namespace)

    def attribute_value(self, name):
        if name in self._attributes:
            return self._attributes[name]
        return None

    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        self._attributes = attributes

    def child_index(self, child):
        """
        Returns the child index value of where it is found in the children
        list

        Args:
            child: Node

        Returns:
            Int index value

        """
        index = None
        try:
            index = self._children.index(child)
        except ValueError as e:
            logger.error(e)
        return index

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, children):
        self._children = children

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        self._content = None if content is None else str(content)

    def copy(self):
        """
        Returns a deep copy (including all children) of the node. All nodes are given new node IDs.

        Returns:
            Node

        """
        # Make a shallow copy and give it a new ID
        _copy = copy.copy(self)
        _copy._id = str(uuid.uuid1())
        Node.set_node_instance(_copy)
        # Construct the attributes dictionary so it's not just a reference to self's version
        _copy.attributes = {}
        for key, val in self.attributes.items():
            _copy.attributes[key] = val
        _copy.nsmap = {}
        for key, val in self.nsmap.items():
            _copy.nsmap[key] = val
        _copy.extras = {}
        for key, val in self.extras.items():
            _copy.extras[key] = val
        # Construct the children list so it's not just a reference to self's version
        _copy.children = []
        for child in self.children:
            _child_copy = child.copy()
            _child_copy.parent = _copy
            _copy.children.append(_child_copy)
        return _copy

    @property
    def extras(self):
        return self._extras

    @extras.setter
    def extras(self, e: dict):
        self._extras = e

    def find_all_children(self, child_name):
        """
        Returns a list of all children that matches the child_name, or returns
        an empty list if there are no matches.

        Args:
            child_name: Child name to be matched

        Returns:
            List
        """
        return [
            child_node
            for child_node in self._children
            if child_node.name == child_name
        ]

# TODO: remove descendants argument to avoid side-effect and return list object
    def find_all_descendants(self, child_name, descendants):
        """
        Generates a list of all descendants that match the child_name, or
        an empty list if there are no matches.

        Args:
            child_name: Child name to be matched
            descendants: List object to be filled with descendant nodes
        """
        for child_node in self._children:
            if child_node.name == child_name:
                descendants.append(child_node)
            child_node.find_all_descendants(child_name, descendants)

    @property
    def id(self):
        """
        Returns the unique identifier of the node instance
        Returns:
            Str
        """
        return self._id

    def list_attributes(self):
        return list(self._attributes.keys())

    def find_descendant(self, descendant_name):
        """
        Recursively searches for the first descendant that matches the
        descendant_name and returns it, or returns None if there is no
        match.

        Args:
            descendant_name: Descendant node name to be matched

        Returns
            Node or None
        """
        descendant = None
        for descendant_node in self._children:
            if descendant_node.name == descendant_name:
                descendant = descendant_node
            else:
                descendant = descendant_node.find_descendant(descendant_name=descendant_name)
            if descendant:
                break
        return descendant

    def find_child(self, child_name):
        """
        Searches for the first child that matches the
        child_name and returns it, or returns None if there is no
        match. Does not search beyond the first generation of descendants.

        Args:
            child_name: Child name to be matched

        Returns
            Node or None
        """
        for child_node in self._children:
            if child_node.name == child_name:
                return child_node
        return None

    def find_single_node_by_path(self, path: list):
        """
        Search down a descendant lineage, using the names in the path provided. Return
        the first node found that satisfies the path.

        To give an example based on EML 2.2, if self is the EML node, then the path
        [names.DATASET, names.CREATOR, names.USERID] will find the userID node of the
        first creator of the dataset, if any. Note that there may be many userID nodes in the
        tree (for creators, metadataProviders, associatedParties, project personnel, etc.,
        so just doing a recursive search for a userID node isn't likely to return the
        desired result.

        Note that this method only returns a single node (or None, if none is found). To
        get all nodes satisfying the path, use find_all_nodes_by_path.

        Args:
            path: List of node names defining the descendant lineage to be found.

        Returns
            Node or None
        """
        if not path or len(path) == 0:
            return None
        current_node = self
        for name in path:
            if not current_node:
                return None
            current_node = current_node.find_child(name)
        return current_node

    def find_all_nodes_by_path(self, path: list):
        """
        Search down a descendant lineage, using the names in the path provided. Return
        a list of nodes that satisfy the path.

        To give an example based on EML 2.2, if self is the EML node, then the path
        [names.DATASET, names.CREATOR, names.USERID] will return a list consisting of
        the userIDs for all of the creators of the dataset, if any.

        Note that this method returns a list of nodes (or None, if none is found). To
        get a single node satisfying the path, use find_single_node_by_path. In cases
        where it is known that at most one node can satisfy the path, getting a single
        node is more convenient than getting a list.

        Args:
            path: List of node names defining the descendant lineages to be found.

        Returns
            List of Nodes, which may be empty
        """
        if not path or len(path) == 0:
            return []
        current_list = [self]
        for name in path:
            if not current_list:
                return []
            next_generation = []
            for node in current_list:
                next_generation.extend(node.find_all_children(name))
            current_list = next_generation
        return current_list

    def get_ancestry(self):
        ancestry = []
        node = self
        while True:
            ancestry.insert(0, node)
            node = node.parent
            if not node:
                break
        return ancestry

    @staticmethod
    def is_equal(node1, node2) -> bool:
        if id(node1) == id(node2):
            return False
        if node1.name != node2.name:
            return False
        if node1.content != node2.content:
            return False
        if node1.tail != node2.tail:
            return False
        if len(node1.attributes) != len(node2.attributes):
            return False
        else:
            for key in node1.attributes.keys():
                try:
                    if node1.attributes[key] != node2.attributes[key]:
                        return False
                except KeyError:
                    return False
        if len(node1.nsmap) != len(node2.nsmap):
            return False
        else:
            for key in node1.nsmap.keys():
                try:
                    if node1.nsmap[key] != node2.nsmap[key]:
                        return False
                except KeyError:
                    return False
        if node1.prefix != node2.prefix:
            return False
        if len(node1.extras) != len(node2.extras):
            return False
        else:
            for key in node1.extras.keys():
                try:
                    if node1.extras[key] != node2.extras[key]:
                        return False
                except KeyError:
                    return False
        if len(node1.children) != len(node2.children):
            return False
        else:
            for index in range(len(node1.children)):
                child1 = node1.children[index]
                child2 = node2.children[index]
                return Node.is_equal(child1, child2)
        return True

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def nsmap(self):
        return self._nsmap
    
    @nsmap.setter
    def nsmap(self, nsmap: dict):
        self._nsmap = nsmap

    @property
    def object(self):
        return self.__object()

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = prefix

    def remove_attribute(self, name):
        del self._attributes[name]

    def remove_child(self, child):
        """
        Removes the child object from the children list

        Args:
            child: Node

        Returns:
            None
        """
        self._children.remove(child)

    def remove_children(self):
        self._children = []

    def remove_namespace(self, prefix: str, nsmap_id: int = None) -> None:
        if nsmap_id is None:
            nsmap_id = id(self.nsmap)
        if prefix in self.nsmap:
            self.nsmap = copy.deepcopy(self.nsmap)
            del self.nsmap[prefix]

        for child in self._children:
            if id(child.nsmap) == nsmap_id:
                child.nsmap = self.nsmap
                child.remove_namespace(prefix, nsmap_id=nsmap_id)
            else:
                child.remove_namespace(prefix)

    def replace_child(self, old_child: "Node", new_child: "Node", delete_old: bool = True):
        """
        Replaces the old child with a new child

        Args:
            old_child: Node
            new_child: Node
            delete_old: Boolean

        Returns:
            None
        """
        if new_child.name != old_child.name:
            msg = f'Child type "{new_child.name}" and "{old_child.name}" mismatch'
            raise ValueError(msg)

        new_child.parent = self
        self._children[self._children.index(old_child)] = new_child
        if delete_old:
            Node.delete_node_instance(id=old_child.id)

    def shift(self, child, direction: Shift, sib: bool = True):
        """
        Shifts a child's position either left or right of its current position
        or not at all if already at local edge

        Args:
            child: Node to be swapped
            direction: shift (LEFT, RIGHT)
            sib: shift only within same sibling type

        Returns:
            int of new index location or same if no change
        """
        index = self._children.index(child)
        name = self._children[index].name
        if direction == Shift.RIGHT:
            if sib:
                for sib_index in range(index + 1, len(self._children)):
                    if self._children[sib_index].name == name:
                        self.children[index], self.children[sib_index] = (
                            self.children[sib_index],
                            self.children[index],
                        )
                        index = sib_index
                        break
            else:
                if index <= len(self._children) - 1:
                    self.children[index], self.children[index + 1] = (
                        self.children[index + 1],
                        self.children[index],
                    )
        elif direction == Shift.LEFT:
            if sib:
                for sib_index in range(index - 1, -1, -1):
                    if self._children[sib_index].name == name:
                        self.children[index], self.children[sib_index] = (
                            self.children[sib_index],
                            self.children[index],
                        )
                        index = sib_index
                        break
            else:
                if index > 0:
                    self.children[index], self.children[index - 1] = (
                        self.children[index - 1],
                        self.children[index],
                    )
        else:
            msg = "Expected direction to be either Shift.RIGHT or Shift.LEFT"
            raise ValueError(msg)

        return index

    def set_nsmap(self, nsmap: dict, children: bool = True):
        self.nsmap = nsmap
        if children:
            for child in self.children:
                child.set_nsmap(nsmap, children)

    @property
    def tail(self) -> str:
        return self._tail

    @tail.setter
    def tail(self, content: str):
        self._tail = content



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: names

:Synopsis:

:Author:
    servilla
    costa
    ide

:Created:
    6/18/18
"""
import daiquiri


logger = daiquiri.getLogger("names: " + __name__)


# Named constants for EML metadata element names
ABSTRACT = "abstract"
ACCESS = "access"
ACCURACY = "accuracy"
ACKNOWLEDGEMENTS = "ackknowledgements"
ADDITIONALINFO = "additionalInfo"
ADDITIONALMETADATA = "additionalMetadata"
ADDRESS = "address"
ADMINISTRATIVEAREA = "administrativeArea"
ALLOW = "allow"
ALTERNATEIDENTIFIER = "alternateIdentifier"
ALTITUDEMAXIMUM = "altitudeMaximum"
ALTITUDEMINIMUM = "altitudeMinimum"
ALTITUDEUNITS = "altitudeUnits"
ANNOTATION = "annotation"
ASSOCIATEDPARTY = "associatedParty"
ATTRIBUTE = "attribute"
ATTRIBUTEACCURACYEXPLANATION = "attributeAccuracyExplanation"
ATTRIBUTEACCURACYREPORT = "attributeAccuracyReport"
ATTRIBUTEACCURACYVALUE = "attributeAccuracyValue"
ATTRIBUTEDEFINITION = "attributeDefinition"
ATTRIBUTELABEL = "attributeLabel"
ATTRIBUTELIST = "attributeList"
ATTRIBUTENAME = "attributeName"
ATTRIBUTEORIENTATION = "attributeOrientation"
AUTHENTICATION = "authentication"
AWARD = "award"
AWARDNUMBER = "awardNumber"
AWARDURL = "awardUrl"
BANDGAPBYTES = "bandgapbytes"
BANDROWBYTES = "bandrowbytes"
BEGINDATE = "beginDate"
BINARYRASTERFORMAT = "binaryRasterFormat"
BOUNDINGALTITUDES = "boundingAltitudes"
BOUNDINGCOORDINATES = "boundingCoordinates"
BOUNDS = "bounds"
BYTEORDER = "byteorder"
CALENDARDATE = "calendarDate"
CASESENSITIVE = "caseSensitive"
CHARACTERENCODING = "characterEncoding"
CITATION = "citation"
CITY = "city"
CODE = "code"
CODEDEFINITION = "codeDefinition"
CODEEXPLANATION = "codeExplanation"
CODESETNAME = "codesetName"
CODESETURL = "codesetURL"
COLLAPSEDELIMITERS = "collapseDelimiters"
COMMONNAME = "commonName"
COMPLEX = "complex"
COMPRESSIONMETHOD = "compressionMethod"
CONNECTION = "connection"
CONTACT = "contact"
COVERAGE = "coverage"
COUNTRY = "country"
CREATOR = "creator"
CUSTOMUNIT = "customUnit"
DATAFORMAT = "dataFormat"
DATASET = "dataset"
DATASOURCE = "dataSource"
DATATABLE = "dataTable"
DATETIME = "dateTime"
DATETIMEDOMAIN = "dateTimeDomain"
DATETIMEPRECISION = "dateTimePrecision"
DEFINITION = "definition"
DEFINITIONATTRIBUTEREFERENCE = "definitionAttributeReference"
DELIVERYPOINT = "deliveryPoint"
DENY = "deny"
DESCRIBES = "describes"
DESCRIPTION = "description"
DESCRIPTOR = "descriptor"
DESCRIPTORVALUE = "descriptorValue"
DISTRIBUTION = "distribution"
EASTBOUNDINGCOORDINATE = "eastBoundingCoordinate"
ELECTRONICMAILADDRESS = "electronicMailAddress"
ENTITYCODELIST = "entityCodeList"
ENTITYREFERENCE = "entityReference"
EXTERNALCODESET = "externalCodeSet"
EXTERNALLYDEFINEDFORMAT = "externallyDefinedFormat"
EML = "eml"
EMPHASIS = "emphasis"
ENCODINGMETHOD = "encodingMethod"
ENDDATE = "endDate"
ENTITYDESCRIPTION = "entityDescription"
ENTITYNAME = "entityName"
ENTITYTYPE = "entityType"
ENUMERATEDDOMAIN = "enumeratedDomain"
FIELDDELIMITER = "fieldDelimiter"
FIELDSTARTCOLUMN = "fieldStartColumn"
FIELDWIDTH = "fieldWidth"
FORMATNAME = "formatName"
FORMATSTRING = "formatString"
FORMATVERSION = "formatVersion"
FUNDERIDENTIFIER = "funderIdentifier"
FUNDERNAME = "funderName"
FUNDING = "funding"
GIVENNAME = "givenName"
GENERALTAXONOMICCOVERAGE = "generalTaxonomicCoverage"
GEOGRAPHICCOVERAGE = "geographicCoverage"
GEOGRAPHICDESCRIPTION = "geographicDescription"
GETTINGSTARTED = "gettingStarted"
IDENTIFIER = "identifier"
INDIVIDUALNAME = "individualName"
INLINE = "inline"
INSTRUMENTATION = "instrumentation"
INTELLECTUALRIGHTS = "intellectualRights"
INTERVAL = "interval"
INTRODUCTION = "introduction"
ITEMIZEDLIST = "itemizedlist"
KEYWORD = "keyword"
KEYWORDSET = "keywordSet"
KEYWORDTHESAURUS = "keywordThesaurus"
LANGUAGE = "language"
LAYOUT = "layout"
LICENSED = "licensed"
LICENSENAME = "licenseName"
LINENUMBER = "lineNumber"
LISTITEM = "listitem"
LITERALCHARACTER = "literalCharacter"
LITERALLAYOUT = "literalLayout"
MAINTENANCE = 'maintenance'
MAINTENANCEUPDATEFREQUENCY = 'maintenanceUpdateFrequency'
MARKDOWN = "markdown"
MAXIMUM = "maximum"
MAXRECORDLENGTH = "maxRecordLength"
MEASUREMENTSCALE = "measurementScale"
MEDIUMNAME = "mediumName"
MEDIUMDENSITY = "mediumDensity"
MEDIUMDENSITYUNITS = "mediumDensityUnits"
MEDIUMVOLUME = "mediumVolume"
MEDIUMFORMAT = "mediumFormat"
MEDIUMNOTE = "mediumNote"
METADATA = "metadata"
METADATAPROVIDER = "metadataProvider"
METHODS = "methods"
METHODSTEP = "methodStep"
MISSINGVALUECODE = "missingValueCode"
MINIMUM = "minimum"
MULTIBAND = "multiBand"
NBANDS = "nbands"
NBITS = "nbits"
NOMINAL = "nominal"
NONNUMERICDOMAIN = "nonNumericDomain"
NORTHBOUNDINGCOORDINATE = "northBoundingCoordinate"
NUMBEROFRECORDS = "numberOfRecords"
NUMBERTYPE = "numberType"
NUMERICDOMAIN = "numericDomain"
NUMFOOTERLINES = "numFooterLines"
NUMHEADERLINES = "numHeaderLines"
NUMPHYSICALLINESPERRECORD = "numPhysicalLinesPerRecord"
OBJECTNAME = "objectName"
OFFLINE = "offline"
ONLINE = "online"
ONLINEDESCRIPTION = "onlineDescription"
ONLINEURL = "onlineUrl"
ORDERATTRIBUTEREFERENCE = "orderAttributeReference"
ORDEREDLIST = "orderedlist"
ORDINAL = "ordinal"
ORGANIZATIONNAME = "organizationName"
OTHERENTITY = "otherEntity"
PARA = "para"
PATTERN = "pattern"
PERMISSION = "permission"
PERSONNEL = "personnel"
PHONE = "phone"
PHYSICAL = "physical"
PHYSICALLINEDELIMITER = "physicalLineDelimiter"
POSITIONNAME = "positionName"
POSTALCODE = "postalCode"
PRECISION = "precision"
PRINCIPAL = "principal"
PROJECT = "project"
PROPERTYURI = "propertyURI"
PROVIDER = "provider"
PUBDATE = "pubDate"
PUBLISHER = "publisher"
PUBPLACE = "pubPlace"
PURPOSE = "purpose"
QUALITYCONTROL = "qualityControl"
QUANTITATIVEATTRIBUTEACCURACYASSESSMENT = (
    "quantitativeAttributeAccuracyAssessment"
)
QUOTECHARACTER = "quoteCharacter"
RANGEOFDATES = "rangeOfDates"
RATIO = "ratio"
REFERENCES = "references"
RELATED_PROJECT = "relatedProject"
ROLE = "role"
ROWCOLUMNORIENTATION = "rowColumnOrientation"
RECORDDELIMITER = "recordDelimiter"
SALUTATION = "salutation"
SAMPLING = "sampling"
SAMPLINGDESCRIPTION = "samplingDescription"
SECTION = "section"
SERIES = "series"
SHORTNAME = "shortName"
SIMPLEDELIMITED = "simpleDelimited"
SINGLEDATETIME = "singleDateTime"
SIZE = "size"
SKIPBYTES = "skipbytes"
SOURCE = "source"
SOUTHBOUNDINGCOORDINATE = "southBoundingCoordinate"
STANDARDUNIT = "standardUnit"
STORAGETYPE = "storageType"

# The following is commented out to accommodate any downstream code that does not support studyAreaDescription in EML
# STUDYAREADESCRIPTION = "studyAreaDescription"

STUDYEXTENT = "studyExtent"
SUBSCRIPT = "subscript"
SUPERSCRIPT = "superscript"
SURNAME = "surName"
TAXONID = "taxonId"
TAXONOMICCLASSIFICATION = "taxonomicClassification"
TAXONOMICCOVERAGE = "taxonomicCoverage"
TAXONRANKNAME = "taxonRankName"
TAXONRANKVALUE = "taxonRankValue"
TEXTDELIMITED = "textDelimited"
TEXTDOMAIN = "textDomain"
TEXTFIXED = "textFixed"
TEXTFORMAT = "textFormat"
TEMPORALCOVERAGE = "temporalCoverage"
TIME = "time"
TITLE = "title"
TOTALROWBYTES = "totalrowbytes"
TYPESYSTEM = "typeSystem"
UNIT = "unit"
UNITLIST = "unitList"
URL = "url"
USERID = "userId"
VALUE = "value"
VALUEATTRIBUTEREFERENCE = "valueAttributeReference"
VALUEURI = "valueURI"
WESTBOUNDINGCOORDINATE = "westBoundingCoordinate"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: evaluate.py

:Synopsis:

:Author:
    ide

:Created:
    6/23/20
"""
from enum import Enum, auto


class ValidationError(Enum):
    ATTRIBUTE_EXPECTED_ENUM = auto()
    ATTRIBUTE_REQUIRED = auto()
    ATTRIBUTE_UNRECOGNIZED = auto()
    CHILD_NOT_ALLOWED = auto()
    CONTENT_EXPECTED_EMPTY = auto()
    CONTENT_EXPECTED_ENUM = auto()
    CONTENT_EXPECTED_FLOAT = auto()
    CONTENT_EXPECTED_NONEMPTY = auto()
    CONTENT_EXPECTED_RANGE = auto()
    CONTENT_EXPECTED_STRING = auto()
    CONTENT_EXPECTED_TIME_FORMAT = auto()
    CONTENT_EXPECTED_URI = auto()
    CONTENT_EXPECTED_YEAR_FORMAT = auto()
    MAX_CHOICE_EXCEEDED = auto()
    MAX_OCCURRENCE_EXCEEDED = auto()
    MIN_CHOICE_UNMET = auto()
    MIN_OCCURRENCE_UNMET = auto()
    STR_CONTENT_UNICODE_ERROR = auto()
    UNKNOWN_ATTRIBUTE = auto()
    UNKNOWN_CONTENT_RULE = auto()
    UNKNOWN_NODE = auto()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: rule

:Synopsis:
    EML node rules processing

:Author:
    servilla
    costa
    ide

:Created:
    7/10/18
"""
import datetime
from datetime import time
import importlib.resources
import json
from typing import Optional

import daiquiri
from rfc3986 import uri_reference, validators
from rfc3986.exceptions import (
    InvalidComponentsError,
    MissingComponentError,
    UnpermittedComponentError
)

from metapype.eml.exceptions import (
    ChildNotAllowedError,
    MaxOccurrenceExceededError,
    MetapypeRuleError,
    MinOccurrenceUnmetError,
    ContentExpectedUriError,
    StrContentUnicodeError,
    UnknownContentRuleError,
)
from metapype.eml import names
from metapype.eml.validation_errors import ValidationError
from metapype.model.node import Node


logger = daiquiri.getLogger("validate: " + __name__)


REQUIRED = True
OPTIONAL = False
INFINITY = None

TYPE_NONE = None
TYPE_STR = "str"
TYPE_INT = "int"
TYPE_FLOAT = "float"
TYPE_DATETIME = "datetime"
TYPE_YEARDATE = "yearDate"


def load_rules():
    """
    Load rules from the JSON file into the rules dict
    """

    rules = importlib.resources.read_text("metapype.eml", "rules.json")
    _rules_dict = json.loads(rules)
    return _rules_dict


rules_dict = load_rules()


class Rule(object):
    """
    The Rule class holds rule content for a specific rule as well as the logic for
    processing content validation.
    """

    def __init__(self, rule_name):
        self._name = rule_name
        rule_data = rules_dict[rule_name]
        self._attributes = rule_data[0]
        self._children = rule_data[1]
        self._content = rule_data[2]
        self._rule_children_names = self._get_rule_children_names(self._children)
        self._depth = 0

    @staticmethod
    def child_list_node_names(child_list: list):
        if list is None or len(child_list) < 3:
            raise MetapypeRuleError(
                "Child list must contain at least 3 elements"
            )
        node_names = child_list[:-2]
        return node_names

    @staticmethod
    def child_list_min_occurrences(child_list: list):
        if list is None or len(child_list) < 3:
            raise MetapypeRuleError(
                "Child list must contain at least 3 elements"
            )
        min_occurrences = child_list[-2]
        return min_occurrences

    @staticmethod
    def child_list_max_occurrences(child_list: list):
        if list is None or len(child_list) < 3:
            raise MetapypeRuleError(
                "Child list must contain at least 3 elements"
            )
        max_occurrences = child_list[-1]
        return max_occurrences

    @staticmethod
    def is_float(val: str = None):
        """
        Boolean to determine whether node content is
        (or can be converted to) a valid float value.
        """
        is_valid = False
        if val is None:
            return False
        try:
            __ = float(val)
            is_valid = True
        except ValueError:
            pass
        return is_valid

    @staticmethod
    def is_int(val: str = None):
        """
        Boolean to determine whether node content is
        (or can be converted to) a valid int value.
        """
        is_valid = False
        if val:
            try:
                __ = int(val)
                is_valid = True
            except ValueError:
                pass
        return is_valid

    @staticmethod
    def is_yeardate(val: str = None):
        """
        Boolean to determine whether node content is a valid yearDate value.
        """
        is_valid = False
        if val and type(val) is str:
            for yeardate_format in ["%Y", "%Y-%m-%d"]:
                try:
                    datetime.datetime.strptime(val, yeardate_format)
                    is_valid = True
                    break
                except ValueError as ex:
                    logger.debug(ex)
        return is_valid

    @staticmethod
    def is_uri(val: str = None) -> bool:
        """
        Boolean to determine whether node content is a valid uri.
        Args:
            val: String value of uri

        Returns: boolean

        """
        is_valid = False
        validator = validators.Validator().allow_schemes(
            "http", "https", "ftp"
        ).require_presence_of(
            "scheme", "host"
        ).check_validity_of(
            "scheme", "host", "path"
        )
        uri = uri_reference(val)
        try:
            validator.validate(uri)
            is_valid = True
        except (InvalidComponentsError, MissingComponentError, UnpermittedComponentError) as ex:
            logger.debug(ex)
        return is_valid

    @staticmethod
    def is_time(val: str = None):
        """
        Boolean to determine whether node content is a valid time value.
        """
        is_valid = False
        if val and type(val) is str:
            try:
                time.fromisoformat(val)
                is_valid = True
            except ValueError as ex:
                logger.debug(ex)
        return is_valid

    def child_insert_index(self, parent: Node, new_child: Node) -> int:
        """
        Determines the index location of a new child node in a given parent node
        based on the parent node's rule type. Note that only a legal position in
        the list of possible children is guaranteed; parent node validity may be
        based on other constraint of the rule type.

        Args:
            parent: Parent node of which to be adding child node
            new_child: Child node to be added

        Returns:
            int: Index location of new child node

        """
        try:
            new_child_index = self._rule_children_names.index(new_child.name)
        except ValueError as e:
            msg = f"Child '{new_child.name}' not allowed in parent '{parent.name}'"
            raise ChildNotAllowedError(msg)
        for index, child in enumerate(parent.children):
            parent_child_index = self._rule_children_names.index(child.name)
            if parent_child_index > new_child_index:
                return index
        index = len(parent.children)
        return index

    @staticmethod
    def _is_in_path(rule_children: list, node: Node) -> bool:
        is_in_path = False
        if Rule._is_rule_child(rule_children):
            return rule_children[0] == node.name
        elif Rule._is_sequence(rule_children):
            for child in rule_children:
                is_in_path = Rule._is_in_path(child, node)
                if is_in_path:
                    break
        else:
            for child in rule_children[:-2]:
                is_in_path = Rule._is_in_path(child, node)
                if is_in_path:
                    break
        return is_in_path

    def is_required_attribute(self, attribute: str):
        if attribute in self._attributes:
            return self._attributes[attribute][0]
        else:
            raise Exception(f"Unknown attribute {attribute}")

    def is_allowed_child(self, child_name: str):
        return child_name in self._rule_children_names

    def allowed_attribute_values(self, attribute: str):
        values = []
        if attribute in self._attributes:
            if len(self._attributes[attribute]) > 1:
                values = self._attributes[attribute][1:]
        else:
            raise Exception(f"Unknown attribute {attribute}")
        return values

    def has_enum_content(self):
        return "content_enum" in self._content

    def validate_rule(self, node: Node, errs: list = None):
        """
        Validates a node for rule compliance by validating the node's
        content, attributes, and children.

        Args:
            node: Node instance to be validated
            errs: List of validation errors

        Returns:
            None

        Raises:
            MetapypeRuleError: Illegal attribute or missing required attribute
        """
        if self.name in (RULE_TEXT, RULE_ANYNAME, RULE_PARA, RULE_SUBSCRIPT, RULE_SUPERSCRIPT):
            is_mixed_content = True
        else:
            is_mixed_content = False

        self._validate_content(node, is_mixed_content, errs)
        self._validate_attributes(node, errs)
        self._validate_children(node, is_mixed_content, errs)

    def _validate_content(self, node: Node, is_mixed_content: bool, errs: list = None):
        """
        Validates node content for rule compliance.
        For each of the content rules configured for this rule,
        validates the node to see if its content complies
        with the content that this rule expects.

        Args:
            node: Node instance to be validated

        Returns:
            None

        Raises:
            MetapypeRuleError: Illegal attribute or missing required attribute
        """
        for content_rule in self._content["content_rules"]:
            if content_rule == "emptyContent":
                self._validate_empty_content(node, errs)
            elif content_rule == "floatContent":
                self._validate_float_content(node, errs)
            elif content_rule == "floatRangeContent_EW":
                self._validate_float_range_ew_content(node, errs)
            elif content_rule == "floatRangeContent_NS":
                self._validate_float_range_ns_content(node, errs)
            elif content_rule == "floatContent_Nonnegative":
                self._validate_float_content_nonnegative(node, errs)
            elif content_rule == "intContent":
                self._validate_int_content(node, errs)
            elif content_rule == "nonEmptyContent":
                self._validate_non_empty_content(node, is_mixed_content, errs)
            elif content_rule == "strContent":
                self._validate_str_content(node, errs)
            elif content_rule == "timeContent":
                self._validate_time_content(node, errs)
            elif content_rule == 'uriContent':
                self._validate_uri_content(node, errs)
            elif content_rule == "yearDateContent":
                self._validate_yeardate_content(node, errs)
            elif content_rule == "anyContent":
                pass
            else:
                msg = f"Node {node.name} content type rule {content_rule} not recognized"
                if errs is None:
                    raise UnknownContentRuleError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.UNKNOWN_CONTENT_RULE,
                            msg,
                            node,
                        )
                    )

        if self.has_enum_content():
            enum_values = self._content["content_enum"]
            self._validate_enum_content(node, enum_values, errs)

    @staticmethod
    def _validate_empty_content(node: Node, errs: list = None):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_EMPTY,
                        msg,
                        node,
                        node.content,
                    )
                )

    @staticmethod
    def _validate_enum_content(
        node: Node, enum_values: list, errs: list = None
    ):
        if node.content not in enum_values:
            msg = f'Node "{node.name}" content should be one of "{enum_values}", not "{node.content}"'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_ENUM,
                        msg,
                        node,
                        enum_values,
                        node.content,
                    )
                )

    @staticmethod
    def _validate_int_content(node: Node, errs: list = None):
        val = node.content
        if val is not None and not Rule.is_int(val):
            msg = f'Node "{node.name}" content should be type "{TYPE_INT}", not "{type(node.content)}"'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_INT,
                        msg,
                        node,
                        type(node.content),
                    )
                )

    @staticmethod
    def _validate_float_content(node: Node, errs: list = None):
        val = node.content
        if val is not None and not Rule.is_float(val):
            msg = f'Node "{node.name}" content should be type "{TYPE_FLOAT}", not "{type(node.content)}"'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_FLOAT,
                        msg,
                        node,
                        type(node.content),
                    )
                )

    def _validate_float_range_content(self, node: Node, minmax, errs: list = None):
        self._validate_float_content(node, errs)
        float_val = float(node.content)
        if float_val < minmax[0] or float_val > minmax[1]:
            msg = f'Node "{node.name}" content should be in range {minmax}'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_RANGE,
                        msg,
                        node,
                        minmax[0],
                        minmax[1],
                        float_val,
                    )
                )

    def _validate_float_nonnegative(
        self, node: Node, errs: list = None
    ):
        self._validate_float_content(node, errs)
        float_val = float(node.content)
        if float_val < 0:
            msg = f'Node "{node.name}" content should be non-negative'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_RANGE,
                        msg,
                        node,
                        0,
                        None,
                        float_val,
                    )
                )

    def _validate_float_range_ew_content(self, node: Node, errs: list = None):
        self._validate_float_range_content(node, (-180.0, 180.0), errs)

    def _validate_float_range_ns_content(self, node: Node, errs: list = None):
        self._validate_float_range_content(node, (-90.0, 90.0), errs)

    def _validate_float_content_nonnegative(self, node: Node, errs: list = None):
        self._validate_float_nonnegative(node, errs)

    @staticmethod
    def _validate_non_empty_content(node: Node, is_mixed_content: bool, errs: list = None):
        if node.content is None or not str(node.content).strip():  # If purely whitespace, it's considered empty
            if (is_mixed_content and len(node.children) == 0) or not is_mixed_content:
                msg = f'Node "{node.name}" content should not be empty or entirely whitespace'
                if errs is None:
                    raise MetapypeRuleError(msg)
                else:
                    errs.append(
                        (ValidationError.CONTENT_EXPECTED_NONEMPTY, msg, node)
                    )

    @staticmethod
    def _validate_str_content(node: Node, errs: list = None):
        if node.content is not None and type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_STRING,
                        msg,
                        node,
                        type(node.content),
                    )
                )
        if node.content is not None:
            try:
                node.content.encode(encoding="utf-8", errors="strict")
            except UnicodeError as ex:
                msg = f'Node "{node.name}" content contains non-unicode character(s)'
                if errs is None:
                    raise StrContentUnicodeError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.CONTENT_EXPECTED_STRING,
                            msg,
                            node,
                            type(node.content),
                        )
                    )

    @staticmethod
    def _validate_time_content(node: Node, errs: list = None):
        val = node.content
        if val is not None and not Rule.is_time(val):
            msg = f'Node "{node.name}" format should be time ("HH:MM:SS" or "HH:MM:SS.f")'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_TIME_FORMAT,
                        msg,
                        node,
                        node.content,
                    )
                )

    @staticmethod
    def _validate_uri_content(node: Node, errs: list = None):
        uri = node.content
        if uri is not None and not Rule.is_uri(uri):
            msg = f'Node "{node.name}" uri content "{uri}" is not valid'
            if errs is None:
                raise ContentExpectedUriError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_URI,
                        msg,
                        node,
                        node.content,
                    )
                )

    @staticmethod
    def _validate_yeardate_content(node: Node, errs: list = None):
        val = node.content
        if val is not None and not Rule.is_yeardate(val):
            msg = f'Node "{node.name}" format should be year ("YYYY") or date ("YYYY-MM-DD")'
            if errs is None:
                raise MetapypeRuleError(msg)
            else:
                errs.append(
                    (
                        ValidationError.CONTENT_EXPECTED_YEAR_FORMAT,
                        msg,
                        node,
                        node.content,
                    )
                )

    def _validate_attributes(self, node: Node, errs: list = None) -> None:
        """
        Validates node attributes for rule compliance.

        Iterates through the dict of attribute rules and validates whether
        the node instance complies with the rule.

        Args:
            node: Node instance to be validated

        Returns:
            None

        Raises:
            MetapypeRuleError: Illegal attribute or missing required attribute
        """
        for attribute in self._attributes:
            required = self._attributes[attribute][0]
            # Test for required attributes
            if required and attribute not in node.attributes:
                msg = f'"{attribute}" is a required attribute of node "{node.name}"'
                if errs is None:
                    raise MetapypeRuleError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.ATTRIBUTE_REQUIRED,
                            msg,
                            node,
                            attribute,
                        )
                    )
        for attribute in node.attributes:
            # Test for non-allowed attribute
            if attribute not in self._attributes:
                msg = f'"{attribute}" is not a recognized attribute of node "{node.name}"'
                if errs is None:
                    raise MetapypeRuleError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.ATTRIBUTE_UNRECOGNIZED,
                            msg,
                            node,
                            attribute,
                        )
                    )
            else:
                # Test for enumerated list of allowed values
                if (
                    len(self._attributes[attribute]) > 1
                    and node.attributes[attribute]
                    not in self._attributes[attribute][1:]
                ):
                    msg = f'Node "{node.name}" attribute "{attribute}" must be one of the following: "{self._attributes[attribute][1:]}"'
                    if errs is None:
                        raise MetapypeRuleError(msg)
                    else:
                        errs.append(
                            (
                                ValidationError.ATTRIBUTE_EXPECTED_ENUM,
                                msg,
                                node,
                                attribute,
                                self._attributes[attribute][1:],
                            )
                        )

    def _validate_children(self, node: Node, is_mixed_content: bool, errs: list = None) -> None:
        """
        Validates node children for rule compliance.

        1. Ignores validation of children if parent node is "metadata"
        2. Ensures children are valid for node
        3. Iterates through the list children rules and validates whether
           the node instance complies with the rules.

        Returns:
            None

        Raises:
            MetapypeRuleError: Illegal child, bad sequence or choice, missing
            child, or wrong child cardinality
        """
        self._node = node
        self._node_children_names = []
        self._node_index = 0

        if self._node.name == names.METADATA:
            # Metadata nodes may contain any type of child node, but only one such node
            if len(self._node.children) > 1:
                msg = f"Maximum occurrence of 1 child exceeded in parent '{names.METADATA}'"
                if errs is None:
                    raise MaxOccurrenceExceededError(msg)
                else:
                    errs.append((ValidationError.MAX_OCCURRENCE_EXCEEDED, msg, self._node))

        else:
            for node_child in self._node.children:
                self._node_children_names.append(node_child.name)

            # Test for non-valid children
            for node_child_name in self._node_children_names:
                if not self.is_allowed_child(node_child_name):
                    msg = f"Child '{node_child_name}' not allowed in parent '{self._node.name}'"
                    if errs is None:
                        raise ChildNotAllowedError(msg)
                    else:
                        errs.append((ValidationError.CHILD_NOT_ALLOWED, msg, self._node, node_child_name))

            if len(self._children) > 0:
                # Begin validation of children
                modality = Rule._get_children_modality(self._children)
                if modality == "sequence":
                    self._validate_sequence(self._children, is_mixed_content, errs)
                else:
                    self._validate_choice(self._children, is_mixed_content, errs)

            if self._node_index != len(self._node_children_names):
                msg = (
                    f"Child '{self._node_children_names[self._node_index]}' "
                    f"is not allowed in this position for parent '{self._node.name}'"
                )
                if errs is None:
                    raise ChildNotAllowedError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.CHILD_NOT_ALLOWED,
                            msg,
                            self._node,
                            self._node_children_names[self._node_index],
                        )
                    )

    def _validate_sequence(self, rule_children: list, is_mixed_content: bool, errs: list = None):
        for rule_child in rule_children:
            if Rule._get_children_modality(rule_child) == "choice":
                self._validate_choice(rule_child, is_mixed_content, errs)
            else:
                self._validate_rule_child(rule_child, False, errs)

    def _validate_choice(self, rule_children: list, is_mixed_content: bool, errs: list = None):
        choice_min = rule_children[-2]
        choice_max = rule_children[-1]
        choice_occurrence = 0

        while (
                self._node_index < len(self._node_children_names) and
                self._node_children_names[self._node_index] in self._get_rule_children_names(rule_children)
        ):
            for rule_child in rule_children[:-2]:
                if self._node_index == len(self._node_children_names):
                    break
                modality = Rule._get_children_modality(rule_child)
                if modality == "sequence":
                    if self._node_children_names[self._node_index] in self._get_rule_children_names(rule_child):
                        self._validate_sequence(rule_child, is_mixed_content, errs)
                        choice_occurrence += 1
                elif modality == "choice":
                    if self._node_children_names[self._node_index] in self._get_rule_children_names(rule_child):
                        self._validate_choice(rule_child, is_mixed_content, errs)
                        choice_occurrence += 1
                else:
                    if self._node_children_names[self._node_index] == rule_child[0]:
                        self._validate_rule_child(rule_child, True, errs)
                        choice_occurrence += 1
        if choice_max is not INFINITY and choice_occurrence > choice_max:
            msg = f"Maximum occurrence of '{choice_max}' exceeded for choice in parent '{self._node.name}'"
            if errs is None:
                raise MaxOccurrenceExceededError(msg)
            else:
                errs.append(
                    (
                        ValidationError.MAX_CHOICE_EXCEEDED,
                        msg,
                        self._node,
                        self._node.name,
                        choice_max,
                    )
                )
        if choice_occurrence < choice_min and not is_mixed_content:
            msg = f"Minimum occurrence of '{choice_min}' not met for choice in parent '{self._node.name}'"
            if errs is None:
                raise MinOccurrenceUnmetError(msg)
            else:
                errs.append(
                    (
                        ValidationError.MIN_CHOICE_UNMET,
                        msg,
                        self._node,
                        self._node.name,
                        choice_min,
                    )
                )

    def _validate_rule_child(self, rule_child: list, limit_max: bool, errs: list = None):
        rule_child_name = rule_child[0]
        rule_child_min = rule_child[-2]
        rule_child_max = rule_child[-1]
        occurrence = 0
        while (
                self._node_index < len(self._node_children_names) and
                rule_child_name == self._node_children_names[self._node_index]
        ):
            occurrence += 1
            self._node_index += 1
            if limit_max and occurrence == rule_child_max:
                return None
            if rule_child_max is not INFINITY and occurrence > rule_child_max:
                msg = (
                    f"Maximum occurrence of '{rule_child_max}' "
                    f"exceeded for child '{self._node_children_names[self._node_index]}' in parent "
                    f"'{self._node.name}'"
                )
                if errs is None:
                    raise MaxOccurrenceExceededError(msg)
                else:
                    errs.append(
                        (
                            ValidationError.MAX_OCCURRENCE_EXCEEDED,
                            msg,
                            self._node,
                            None if self._node_index >= len(self._node_children_names)
                            else self._node_children_names[self._node_index],
                            rule_child_max,
                        )
                    )
        if occurrence < rule_child_min:
            msg = (
                f"Minimum occurrence of '{rule_child_min}' "
                f"not met for child '{rule_child_name}' in parent '{self._node.name}'"
            )
            if errs is None:
                raise MinOccurrenceUnmetError(msg)
            else:
                errs.append(
                    (
                        ValidationError.MIN_OCCURRENCE_UNMET,
                        msg,
                        self._node,
                        rule_child_name,
                        # None if self._node_index >= len(self._node_children_names)
                        # else self._node_children_names[self._node_index],
                        rule_child_min,
                    )
                )

    @staticmethod
    def _get_children_modality(rule_children: list) -> str:
        if Rule._is_rule_child(rule_children):
            mode = "child_rule"
        elif Rule._is_sequence(rule_children):
            mode = "sequence"
        elif Rule._is_choice(rule_children):
            mode = "choice"
        else:
            msg = f"Unknown modality for {rule_children}"
            raise ValueError(msg)
        return mode

    @staticmethod
    def _is_rule_child(rule_children: list) -> bool:
        # of the form ["str", int, int|None]
        is_rule_child = False
        if isinstance(rule_children[0], str):
            is_rule_child = True
        return is_rule_child

    @staticmethod
    def _is_sequence(rule_children: list) -> bool:
        # of the form [["str", int, int|None]...[]]]
        is_sequence = False
        if len(rule_children) >= 1 and isinstance(rule_children[-1], list):
            is_sequence = True
        return is_sequence

    @staticmethod
    def _is_choice(rule_children: list) -> bool:
        # of the form [["str", int, int|None]...[], int, int|None]]
        is_choice = False
        if len(rule_children) >= 3 and (isinstance(rule_children[0], list) and isinstance(rule_children[-2], int)):
            is_choice = True
        return is_choice

    @staticmethod
    def _get_rule_children_names(children: list) -> list:
        children_names = []
        if len(children) > 0:
            modality = Rule._get_children_modality(children)
            if modality == "choice":
                children_names += Rule._get_rule_children_names(children[:-2])
            elif modality == "sequence":
                for child in children:
                    children_names += Rule._get_rule_children_names(child)
            else:
                children_names.append(children[0])
        return children_names

    @property
    def name(self):
        return self._name

    @property
    def attributes(self):
        return self._attributes

    @property
    def children(self):
        return self._children

    @property
    def content_rules(self):
        return self._content["content_rules"]

    @property
    def content_enum(self):
        if self.has_enum_content():
            return self._content["content_enum"]
        else:
            return []


# Named constants for EML metadata rules
RULE_ACCESS = "accessRule"
RULE_ACCURACY = "accuracyRule"
RULE_ADDITIONALMETADATA = "additionalMetadataRule"
RULE_ADDRESS = "addressRule"
RULE_ALLOW = "allowRule"
RULE_ALTERNATEIDENTIFIER = "alternateIdentifierRule"
RULE_ALTITUDEUNITS = "altitudeUnitsRule"
RULE_ANNOTATION = "annotationRule"
RULE_ANYFLOAT = "anyFloatRule"
RULE_ANYINT = "anyIntRule"
RULE_ANYNAME = "anyNameRule"
RULE_ANYNONNEGATIVEFLOAT = "anyNonNegativeFloatRule"
RULE_ANYSTRING = "anyStringRule"
RULE_ANYURI = "anyURIRule"
RULE_ATTRIBUTE = "attributeRule"
RULE_ATTRIBUTELIST = "attributeListRule"
RULE_AUTHENTICATION = "authenticationRule"
RULE_AWARD = "awardRule"
RULE_BINARYRASTER_FORMAT = "binaryRasterFormatRule"
RULE_BOUNDINGALTITUDES = "boundingAltitudesRule"
RULE_BOUNDINGCOORDINATE_EW = "boundingCoordinateRule_EW"
RULE_BOUNDINGCOORDINATE_NS = "boundingCoordinateRule_NS"
RULE_BOUNDINGCOORDINATES = "boundingCoordinatesRule"
RULE_BOUNDS = "boundsRule"
RULE_CODEDEFINITION = "codeDefinitionRule"
RULE_COMPLEX = "complexRule"
RULE_COVERAGE = "coverageRule"
RULE_DATAFORMAT = "dataFormatRule"
RULE_DATASET = "datasetRule"
RULE_DATASOURCE = "datasetRule"
RULE_DATATABLE = "dataTableRule"
RULE_DATETIME = "dateTimeRule"
RULE_DATETIMEDOMAIN = "dateTimeDomainRule"
RULE_DENY = "denyRule"
RULE_DESCRIPTOR = "descriptorRule"
RULE_DISTRIBUTION = "distributionRule"
RULE_EML = "emlRule"
RULE_EMPHASIS = "anyStringRule"
RULE_ENTITYCODELIST = "entityCodeListRule"
RULE_ENUMERATEDDOMAIN = "enumeratedDomainRule"
RULE_EXTERNALCODESET = "externalCodeSetRule"
RULE_EXTERNALLYDEFINIEDFORMAT = "externallyDefinedFormatRule"
RULE_FUNDING = "fundingRule"
RULE_GEOGRAPHICCOVERAGE = "geographicCoverageRule"
RULE_INDIVIDUALNAME = "individualNameRule"
RULE_INTERVALRATIO = "intervalRatioRule"
RULE_ITEMIZEDLIST = "itemizedlistRule"
RULE_KEYWORD = "keywordRule"
RULE_KEYWORDSET = "keywordSetRule"
RULE_KEYWORDTHESAURUS = "keywordThesaurusRule"
RULE_LICENSED = "licensedRule"
RULE_LISTITEM = "listitemRule"
RULE_MAINTENANCE = "maintenanceRule"
RULE_MEASUREMENTSCALE = "measurementScaleRule"
RULE_METADATA = "metadataRule"
RULE_METHODS = "methodsRule"
RULE_METHODSTEP = "methodStepRule"
RULE_MINMAX = "minMaxRule"
RULE_MISSINGVALUECODE = "missingValueCodeRule"
RULE_MULTIBAND = "multiBandRule"
RULE_NOMINAL = "nominalOrdinalRule"
RULE_NONNUMERICDOMAIN = "nonNumericDomainRule"
RULE_NUMERICDOMAIN = "numericDomainRule"
RULE_OFFLINE = "offlineRule"
RULE_ONLINE = "onlineRule"
RULE_ORDEREDLIST = "orderedlistRule"
RULE_ORDINAL = "nominalOrdinalRule"
RULE_OTHERENTITY = "otherEntityRule"
RULE_PARA = "paraRule"
RULE_PERMISSION = "permissionRule"
RULE_PHONE = "phoneRule"
RULE_PHYSICAL = "physicalRule"
RULE_PRINCIPAL = "principalRule"
RULE_PROJECT = "projectRule"
RULE_PROPERTYURI = "propertyUriRule"
RULE_QUALITYCONTROL = "qualityControlRule"
RULE_QUANTITATIVEATTRIBUTEACCURACYASSESSMENT = ("quantitativeAttributeAccuracyAssessmentRule")
RULE_RANGEOFDATES = "rangeOfDatesRule"
RULE_RATIO = "ratioRule"
RULE_REFERENCES = "referencesRule"
RULE_RELATEDPROJECT = "relatedProjectRule"
RULE_RESPONSIBLEPARTY = "responsiblePartyRule"
RULE_RESPONSIBLEPARTY_WITH_ROLE = "responsiblePartyWithRoleRule"
RULE_ROLE = "roleRule"
RULE_ROWCOLUMN = "rowColumnRule"
RULE_SAMPLING = "samplingRule"
RULE_SECTION = "sectionRule"
RULE_SINGLEDATETIME = "singleDateTimeRule"
RULE_SIMPLEDELIMITED = "simpleDelimitedRule"
RULE_SIZE = "sizeRule"
RULE_STORAGETYPE = "storageTypeRule"
RULE_STUDYAREADESCRIPTION = "studyAreaDescriptionRule"
RULE_STUDYEXTENT = "studyExtentRule"
RULE_SUBSCRIPT = "subscriptRule"
RULE_SUPERSCRIPT = "superscriptRule"
RULE_TAXONID = "taxonIdRule"
RULE_TAXONOMICCLASSIFICATION = "taxonomicClassificationRule"
RULE_TAXONOMICCOVERAGE = "taxonomicCoverageRule"
RULE_TEMPORALCOVERAGE = "temporalCoverageRule"
RULE_TEXT = "textRule"
RULE_TEXTDELIMITED = "textDelimitedRule"
RULE_TEXTDOMAIN = "textDomainRule"
RULE_TEXTFIXED = "textFixedRule"
RULE_TEXTFORMAT = "textFormatRule"
RULE_TIME = "timeRule"
RULE_UNIT = "unitRule"
RULE_URL = "urlRule"
RULE_USERID = "userIdRule"
RULE_VALUE = "valueRule"
RULE_VALUEURI = "valueUriRule"
RULE_YEARDATE = "yearDateRule"


# Maps node names to their corresponding metadata rule names
node_mappings = {
    names.ABSTRACT: RULE_TEXT,
    names.ACCESS: RULE_ACCESS,
    names.ACCURACY: RULE_ACCURACY,
    names.ACKNOWLEDGEMENTS: RULE_TEXT,
    names.ADDITIONALINFO: RULE_TEXT,
    names.ADDITIONALMETADATA: RULE_ADDITIONALMETADATA,
    names.ADDRESS: RULE_ADDRESS,
    names.ADMINISTRATIVEAREA: RULE_ANYNAME,
    names.ALLOW: RULE_ALLOW,
    names.ALTERNATEIDENTIFIER: RULE_ALTERNATEIDENTIFIER,
    names.ALTITUDEMAXIMUM: RULE_ANYSTRING,
    names.ALTITUDEMINIMUM: RULE_ANYSTRING,
    names.ALTITUDEUNITS: RULE_ALTITUDEUNITS,
    names.ANNOTATION: RULE_ANNOTATION,
    names.ASSOCIATEDPARTY: RULE_RESPONSIBLEPARTY_WITH_ROLE,
    names.ATTRIBUTE: RULE_ATTRIBUTE,
    names.ATTRIBUTEACCURACYEXPLANATION: RULE_ANYSTRING,
    names.ATTRIBUTEACCURACYREPORT: RULE_ANYSTRING,
    names.ATTRIBUTEACCURACYVALUE: RULE_ANYSTRING,
    names.ATTRIBUTEDEFINITION: RULE_ANYSTRING,
    names.ATTRIBUTELABEL: RULE_ANYSTRING,
    names.ATTRIBUTELIST: RULE_ATTRIBUTELIST,
    names.ATTRIBUTENAME: RULE_ANYSTRING,
    names.ATTRIBUTEORIENTATION: RULE_ANYSTRING,
    names.AUTHENTICATION: RULE_AUTHENTICATION,
    names.AWARD: RULE_AWARD,
    names.AWARDNUMBER: RULE_ANYNAME,
    names.AWARDURL: RULE_ANYNAME,
    names.BANDGAPBYTES: RULE_ANYSTRING,
    names.BANDROWBYTES: RULE_ANYSTRING,
    names.BEGINDATE: RULE_SINGLEDATETIME,
    names.BINARYRASTERFORMAT: RULE_BINARYRASTER_FORMAT,
    names.BOUNDINGALTITUDES: RULE_BOUNDINGALTITUDES,
    names.BOUNDINGCOORDINATES: RULE_BOUNDINGCOORDINATES,
    names.BOUNDS: RULE_BOUNDS,
    names.BYTEORDER: RULE_ANYSTRING,
    names.CALENDARDATE: RULE_YEARDATE,
    names.CASESENSITIVE: RULE_ANYSTRING,
    names.CHARACTERENCODING: RULE_ANYSTRING,
    names.CITATION: RULE_ANYSTRING,
    names.CITY: RULE_ANYNAME,
    names.CODE: RULE_ANYSTRING,
    names.CODEDEFINITION: RULE_CODEDEFINITION,
    names.CODEEXPLANATION: RULE_ANYSTRING,
    names.CODESETNAME: RULE_ANYSTRING,
    names.CODESETURL: RULE_ANYURI,
    names.COMMONNAME: RULE_ANYSTRING,
    names.COLLAPSEDELIMITERS: RULE_ANYSTRING,
    names.COMPLEX: RULE_COMPLEX,
    names.COMPRESSIONMETHOD: RULE_ANYSTRING,
    names.CONNECTION: RULE_ANYSTRING,
    names.CONTACT: RULE_RESPONSIBLEPARTY,
    names.COUNTRY: RULE_ANYNAME,
    names.COVERAGE: RULE_COVERAGE,
    names.CREATOR: RULE_RESPONSIBLEPARTY,
    names.CUSTOMUNIT: RULE_ANYSTRING,
    names.DATAFORMAT: RULE_DATAFORMAT,
    names.DATASET: RULE_DATASET,
    names.DATASOURCE: RULE_DATASOURCE,
    names.DATATABLE: RULE_DATATABLE,
    names.DATETIME: RULE_DATETIME,
    names.DATETIMEDOMAIN: RULE_DATETIMEDOMAIN,
    names.DATETIMEPRECISION: RULE_ANYSTRING,
    names.DEFINITION: RULE_ANYSTRING,
    names.DEFINITIONATTRIBUTEREFERENCE: RULE_ANYSTRING,
    names.DELIVERYPOINT: RULE_ANYNAME,
    names.DENY: RULE_DENY,
    names.DESCRIBES: RULE_ANYSTRING,
    names.DESCRIPTOR: RULE_DESCRIPTOR,
    names.DESCRIPTORVALUE: RULE_ANYSTRING,
    names.DESCRIPTION: RULE_TEXT,
    names.DISTRIBUTION: RULE_DISTRIBUTION,
    names.EASTBOUNDINGCOORDINATE: RULE_BOUNDINGCOORDINATE_EW,
    names.ELECTRONICMAILADDRESS: RULE_ANYNAME,
    names.EML: RULE_EML,
    names.EMPHASIS: RULE_EMPHASIS,
    names.ENCODINGMETHOD: RULE_ANYSTRING,
    names.ENDDATE: RULE_SINGLEDATETIME,
    names.ENTITYCODELIST: RULE_ENTITYCODELIST,
    names.ENTITYDESCRIPTION: RULE_ANYSTRING,
    names.ENTITYNAME: RULE_ANYSTRING,
    names.ENTITYREFERENCE: RULE_ANYSTRING,
    names.ENTITYTYPE: RULE_ANYSTRING,
    names.ENUMERATEDDOMAIN: RULE_ENUMERATEDDOMAIN,
    names.EXTERNALCODESET: RULE_EXTERNALCODESET,
    names.EXTERNALLYDEFINEDFORMAT: RULE_EXTERNALLYDEFINIEDFORMAT,
    names.FIELDDELIMITER: RULE_ANYSTRING,
    names.FIELDSTARTCOLUMN: RULE_ANYSTRING,
    names.FIELDWIDTH: RULE_ANYSTRING,
    names.FORMATNAME: RULE_ANYSTRING,
    names.FORMATSTRING: RULE_ANYSTRING,
    names.FORMATVERSION: RULE_ANYSTRING,
    names.FUNDERIDENTIFIER: RULE_ANYNAME,
    names.FUNDERNAME: RULE_ANYNAME,
    names.FUNDING: RULE_TEXT,
    names.GENERALTAXONOMICCOVERAGE: RULE_ANYSTRING,
    names.GIVENNAME: RULE_ANYNAME,
    names.GEOGRAPHICCOVERAGE: RULE_GEOGRAPHICCOVERAGE,
    names.GEOGRAPHICDESCRIPTION: RULE_ANYSTRING,
    names.GETTINGSTARTED: RULE_TEXT,
    names.IDENTIFIER: RULE_ANYSTRING,
    names.INDIVIDUALNAME: RULE_INDIVIDUALNAME,
    names.INLINE: RULE_ANYSTRING,
    names.INSTRUMENTATION: RULE_ANYSTRING,
    names.INTELLECTUALRIGHTS: RULE_TEXT,
    names.INTERVAL: RULE_INTERVALRATIO,
    names.INTRODUCTION: RULE_TEXT,
    names.ITEMIZEDLIST: RULE_ITEMIZEDLIST,
    names.KEYWORD: RULE_KEYWORD,
    names.KEYWORDSET: RULE_KEYWORDSET,
    names.KEYWORDTHESAURUS: RULE_KEYWORDTHESAURUS,
    names.LANGUAGE: RULE_ANYNAME,
    names.LAYOUT: RULE_ANYSTRING,
    names.LICENSED: RULE_LICENSED,
    names.LICENSENAME: RULE_ANYSTRING,
    names.LINENUMBER: RULE_ANYINT,
    names.LISTITEM: RULE_LISTITEM,
    names.LITERALCHARACTER: RULE_ANYSTRING,
    names.LITERALLAYOUT: RULE_ANYSTRING,
    names.MAINTENANCE: RULE_MAINTENANCE,
    names.MAINTENANCEUPDATEFREQUENCY: RULE_ANYSTRING,
    names.MARKDOWN: RULE_ANYSTRING,
    names.MAXIMUM: RULE_MINMAX,
    names.MAXRECORDLENGTH: RULE_ANYINT,
    names.MEASUREMENTSCALE: RULE_MEASUREMENTSCALE,
    names.MEDIUMNAME: RULE_ANYSTRING,
    names.MEDIUMDENSITY: RULE_ANYSTRING,
    names.MEDIUMDENSITYUNITS: RULE_ANYSTRING,
    names.MEDIUMVOLUME: RULE_ANYSTRING,
    names.MEDIUMFORMAT: RULE_ANYSTRING,
    names.MEDIUMNOTE: RULE_ANYSTRING,
    names.METADATA: RULE_METADATA,
    names.METADATAPROVIDER: RULE_RESPONSIBLEPARTY,
    names.METHODS: RULE_METHODS,
    names.METHODSTEP: RULE_METHODSTEP,
    names.MINIMUM: RULE_MINMAX,
    names.MISSINGVALUECODE: RULE_MISSINGVALUECODE,
    names.MULTIBAND: RULE_MULTIBAND,
    names.NBANDS: RULE_ANYINT,
    names.NBITS: RULE_ANYINT,
    names.NOMINAL: RULE_NOMINAL,
    names.NONNUMERICDOMAIN: RULE_NONNUMERICDOMAIN,
    names.NORTHBOUNDINGCOORDINATE: RULE_BOUNDINGCOORDINATE_NS,
    names.NUMBEROFRECORDS: RULE_ANYSTRING,
    names.NUMBERTYPE: RULE_ANYSTRING,
    names.NUMERICDOMAIN: RULE_NUMERICDOMAIN,
    names.NUMFOOTERLINES: RULE_ANYINT,
    names.NUMHEADERLINES: RULE_ANYINT,
    names.NUMPHYSICALLINESPERRECORD: RULE_ANYINT,
    names.OBJECTNAME: RULE_ANYSTRING,
    names.OFFLINE: RULE_OFFLINE,
    names.ONLINE: RULE_ONLINE,
    names.ONLINEDESCRIPTION: RULE_ANYSTRING,
    names.ONLINEURL: RULE_ANYURI,
    names.ORDERATTRIBUTEREFERENCE: RULE_ANYSTRING,
    names.ORDEREDLIST: RULE_ORDEREDLIST,
    names.ORDINAL: RULE_ORDINAL,
    names.ORGANIZATIONNAME: RULE_ANYNAME,
    names.OTHERENTITY: RULE_OTHERENTITY,
    names.PARA: RULE_PARA,
    names.PATTERN: RULE_ANYSTRING,
    names.PERMISSION: RULE_PERMISSION,
    names.PERSONNEL: RULE_RESPONSIBLEPARTY_WITH_ROLE,
    names.PHONE: RULE_PHONE,
    names.PHYSICAL: RULE_PHYSICAL,
    names.PHYSICALLINEDELIMITER: RULE_ANYSTRING,
    names.POSITIONNAME: RULE_ANYNAME,
    names.POSTALCODE: RULE_ANYNAME,
    names.PRECISION: RULE_ANYNONNEGATIVEFLOAT,
    names.PRINCIPAL: RULE_PRINCIPAL,
    names.PROJECT: RULE_PROJECT,
    names.PROPERTYURI: RULE_PROPERTYURI,
    names.PUBDATE: RULE_YEARDATE,
    names.PUBLISHER: RULE_RESPONSIBLEPARTY,
    names.PUBPLACE: RULE_ANYSTRING,
    names.PURPOSE: RULE_TEXT,
    names.QUALITYCONTROL: RULE_QUALITYCONTROL,
    names.QUANTITATIVEATTRIBUTEACCURACYASSESSMENT: RULE_QUANTITATIVEATTRIBUTEACCURACYASSESSMENT,
    names.QUOTECHARACTER: RULE_ANYSTRING,
    names.RANGEOFDATES: RULE_RANGEOFDATES,
    names.RATIO: RULE_INTERVALRATIO,
    names.RECORDDELIMITER: RULE_ANYSTRING,
    names.REFERENCES: RULE_REFERENCES,
    names.RELATED_PROJECT: RULE_RELATEDPROJECT,
    names.ROLE: RULE_ROLE,
    names.ROWCOLUMNORIENTATION: RULE_ROWCOLUMN,
    names.SALUTATION: RULE_ANYNAME,
    names.SAMPLING: RULE_SAMPLING,
    names.SAMPLINGDESCRIPTION: RULE_TEXT,
    names.SECTION: RULE_SECTION,
    names.SERIES: RULE_ANYSTRING,
    names.SHORTNAME: RULE_ANYSTRING,
    names.SIMPLEDELIMITED: RULE_SIMPLEDELIMITED,
    names.SINGLEDATETIME: RULE_SINGLEDATETIME,
    names.SIZE: RULE_SIZE,
    names.SKIPBYTES: RULE_ANYSTRING,
    names.SOURCE: RULE_ANYSTRING,
    names.SOUTHBOUNDINGCOORDINATE: RULE_BOUNDINGCOORDINATE_NS,
    names.STANDARDUNIT: RULE_ANYSTRING,
    names.STORAGETYPE: RULE_STORAGETYPE,

    # The following is commented out to accommodate any downstream code that does not support studyAreaDescription in EML
    # names.STUDYAREADESCRIPTION: RULE_STUDYAREADESCRIPTION,

    names.STUDYEXTENT: RULE_STUDYEXTENT,
    names.SUBSCRIPT: RULE_SUBSCRIPT,
    names.SUPERSCRIPT: RULE_SUPERSCRIPT,
    names.SURNAME: RULE_ANYNAME,
    names.TAXONID: RULE_TAXONID,
    names.TAXONOMICCLASSIFICATION: RULE_TAXONOMICCLASSIFICATION,
    names.TAXONOMICCOVERAGE: RULE_TAXONOMICCOVERAGE,
    names.TAXONRANKNAME: RULE_ANYSTRING,
    names.TAXONRANKVALUE: RULE_ANYSTRING,
    names.TEMPORALCOVERAGE: RULE_TEMPORALCOVERAGE,
    names.TEXTDELIMITED: RULE_TEXTDELIMITED,
    names.TEXTDOMAIN: RULE_TEXTDOMAIN,
    names.TEXTFIXED: RULE_TEXTFIXED,
    names.TEXTFORMAT: RULE_TEXTFORMAT,
    names.TIME: RULE_TIME,
    names.TITLE: RULE_ANYNAME,
    names.TOTALROWBYTES: RULE_ANYSTRING,
    names.UNIT: RULE_UNIT,
    names.URL: RULE_URL,
    names.USERID: RULE_USERID,
    names.VALUE: RULE_VALUE,
    names.VALUEATTRIBUTEREFERENCE: RULE_ANYSTRING,
    names.VALUEURI: RULE_VALUEURI,
    names.WESTBOUNDINGCOORDINATE: RULE_BOUNDINGCOORDINATE_EW,
}


def node_names():
    """
    Helper function.
    Returns a list of all known node names.
    """
    return list(node_mappings.keys())


def get_rule_name(node_name: str):
    """
    Helper function.
    For a given node name, return its corresponding rule name
    """
    return node_mappings.get(node_name)


def get_rule(node_name: str):
    """
    Helper function.
    For a given node name, instantiate its corresponding rule object and return it
    """
    rule_name = get_rule_name(node_name)
    return Rule(rule_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod:
    normalize

:Synopsis:
    Normalize an EML XML document instance or text content, including replacement of non-breaking space characters
    with space characters.

:Author:
    servilla

:Created:
    2/17/24
"""

import daiquiri
from lxml import etree

logger = daiquiri.getLogger(__name__)

normalize_whitespace = \
    """<xsl:stylesheet version="1.0"
                       xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                       >
           <xsl:output omit-xml-declaration="no" indent="yes"/>

           <!-- Template to copy nodes and apply templates to attributes and child nodes -->
           <xsl:template match="@*|node()">
               <xsl:copy>
                   <!-- Apply templates to attributes first to normalize space, then to child nodes -->
                   <xsl:apply-templates select="@*"/>
                   <xsl:apply-templates select="node()"/>
               </xsl:copy>
           </xsl:template>

           <!-- Template for normalizing space in text nodes, with specific exclusions -->
           <xsl:template match="text()[not(ancestor::markup or ancestor::literalLayout or ancestor::objectName or ancestor::attributeName or ancestor::para)]">
               <xsl:value-of select="normalize-space(.)"/>
           </xsl:template>

           <!-- Template to normalize space in attribute values -->
           <xsl:template match="@*">
               <!-- Create a new attribute with the same name but normalized value -->
               <xsl:attribute name="{name()}">
                   <xsl:value-of select="normalize-space(.)"/>
               </xsl:attribute>
           </xsl:template>
       </xsl:stylesheet>"""


def normalize(content: str, is_xml: bool = False) -> str:
    """
    Normalize whitespace within a string, including replacement of non-breaking space characters
    :param content: String content to be normalized
    :param is_xml: Boolean to indicate if content is XML
    :return: Normalized content as a unicode string
    """
    if is_xml:
        xslt = etree.XSLT(etree.XML(normalize_whitespace))
        normalized = str(xslt(etree.XML(content.replace('\xa0', ' ').encode("utf-8"))))
    else:
        words = content.replace('\xa0', ' ').split(" ")
        normalized = " ".join([word.strip() for word in words if word.strip() != ""])
    return normalized



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
:Mod: references

:Synopsis:

:Author:
    servilla

:Created:
    2/23/21
"""
import daiquiri

import metapype.eml.names as names
from metapype.model.node import Node


logger = daiquiri.getLogger(__name__)


def _register_ids(node: Node) -> dict:
    id_register = dict()
    for a, v in node.attributes.items():
        if a == "id":
            id_register[v] = node
    for child in node.children:
        _ = _register_ids(child)
        for k in _.keys():
            if k in id_register:
                msg = f"Duplicate use of ID: '{k}'"
                raise ValueError(msg)
        id_register = {**id_register, **_}
    return id_register


def expand(node: Node):
    references = list()
    node.find_all_descendants(names.REFERENCES, references)
    ids = _register_ids(node)
    for reference in references:
        if reference.content not in ids:
            msg = f"ID not found for REFERENCE '{reference}'"
            raise ValueError(msg)
        source_node = ids[reference.content]
        destination_node = reference.parent
        destination_node.remove_child(reference)
        Node.delete_node_instance(reference.id)
        for source_child in source_node.children:
            source_child_copy = source_child.copy()
            destination_node.add_child(source_child_copy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""":Mod: rules

:Synopsis:
    EML node rule declarations.

:Author:
    servilla
    costa
    ide

:Created:
    6/5/18
"""
import daiquiri

from metapype.eml.exceptions import MetapypeRuleError
from metapype.eml import names
from metapype.model.node import Node


logger = daiquiri.getLogger("validate: " + __name__)


REQUIRED = True
OPTIONAL = False
INFINITY = None

TYPE_NONE = None
TYPE_STR = "str"
TYPE_INT = "int"
TYPE_FLOAT = "float"
TYPE_DATETIME = "datetime"

PERMISSIONS = ("read", "write", "changePermission", "all")


class Rule(object):
    def __init__(self):
        self._attributes = {}
        self._children = []
        self._content = None

    def _validate_content(self, content, node: Node):
        pass

    def validate_rule(self, node: Node):
        self._validate_content(self._content, node)
        validate_attributes(self._attributes, node)
        validate_children(self._children, node)

    @property
    def attributes(self):
        return self._attributes

    @property
    def children(self):
        return self._children

    @property
    def content(self):
        return self._content


# ==================== Begin of rules section ====================


class Accessrule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {
            "id": [OPTIONAL],
            "system": [OPTIONAL],
            "scope": [OPTIONAL],
            "order": [OPTIONAL, "allowFirst", "denyFirst"],
            "authSystem": [REQUIRED],
        }
        self._children = [["allow", "deny", 1, INFINITY]]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class Additionalmetadatarule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {"id": [OPTIONAL]}
        self._children = [["describes", 0, INFINITY], ["metadata", 1, 1]]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class Allowrule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = [
            ["principal", 1, INFINITY],
            ["permission", 1, INFINITY],
        ]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class Anynamerule(Rule):
    """
    Generic rule for names.

    This is a generic rule for evaluating name-based metadata like
    surName, givenName, or salutations.

    Raises:
        MetapypeRuleError: If content is not string or if without child and
        content is empty
    """

    def __init__(self):
        super().__init__()
        self._attributes = {"lang": [OPTIONAL]}
        self._children = [["value", 0, INFINITY]]
        self._content = TYPE_STR

    def _validate_content(self, content, node: Node):
        if node.content is not None and type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            raise MetapypeRuleError(msg)
        if len(node.children) == 0 and node.content is None:
            msg = f'Node "{node.name}" content should not be empty'
            raise MetapypeRuleError(msg)


class DatasetRule(Rule):
    # TODO: complete rule
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = []
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)

    # Temporary override for null rule
    def validate_rule(self, node: Node):
        pass


class DenyRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = [
            ["principal", 1, INFINITY],
            ["permission", 1, INFINITY],
        ]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class EmlRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {
            "packageId": [REQUIRED],
            "system": [REQUIRED],
            "scope": [OPTIONAL],
            "lang": [OPTIONAL],
        }
        self._children = [
            ["access", 0, 1],
            ["dataset", "citation", "software", "protocol", 1, 1],
            ["additionalMetadata", 0, INFINITY],
        ]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class IndividualNameRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = [
            ["salutation", 0, INFINITY],
            ["givenName", 0, INFINITY],
            ["surName", 1, 1],
        ]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class MetadataRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = []
        self._content = TYPE_STR

    def _validate_content(self, content, node: Node):
        if node.content is not None and type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            raise MetapypeRuleError(msg)


class PermissionRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = []
        self._content = TYPE_STR

    def _validate_content(self, content, node: Node):
        if node.content is None or type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            raise MetapypeRuleError(msg)
        if node.content not in PERMISSIONS:
            msg = f'Node "{node.name}" content should be one of "{PERMISSIONS}", not "{node.content}"'
            raise MetapypeRuleError(msg)


class PrincipalRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {}
        self._children = []
        self._content = TYPE_STR

    def _validate_content(self, content, node: Node):
        if node.content is None or type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            raise MetapypeRuleError(msg)


class ResponsiblePartyRule(Rule):
    """
    Generic rule for any responsibleParty type of metadata like creator or
    contact.
    """

    def __init__(self):
        super().__init__()
        self._attributes = {
            "id": [OPTIONAL],
            "system": [OPTIONAL],
            "scope": [OPTIONAL],
        }
        self._children = [
            [
                "individualName",
                "organizationName",
                "positionName",
                1,
                INFINITY,
            ],
            ["address", 0, INFINITY],
            ["phone", 0, INFINITY],
            ["electronicMailAddress", 0, INFINITY],
            ["onlineUrl", 0, INFINITY],
            ["userId", 0, INFINITY],
        ]
        self._content = TYPE_NONE

    def _validate_content(self, content, node: Node):
        if node.content is not None:
            msg = f'Node "{node.name}" content should be empty'
            raise MetapypeRuleError(msg)


class ValueRule(Rule):
    def __init__(self):
        super().__init__()
        self._attributes = {"xml:lang": [REQUIRED]}
        self._children = []
        self._content = TYPE_STR

    def _validate_content(self, content, node: Node):
        if node.content is None or type(node.content) is not str:
            msg = f'Node "{node.name}" content should be type "{TYPE_STR}", not "{type(node.content)}"'
            raise MetapypeRuleError(msg)


# ===================== End of rules section =====================


def validate_attributes(attributes: dict, node: Node) -> None:
    """
    Validates node attributes for rule compliance.

    Iterates through the dict of attribute rules and validates whether
    the node instance complies with the rule.

    Args:
        attributes: dict of rule attributes
        node: Node instance to be validates

    Returns:
        None

    Raises:
        MetapypeRuleError: Illegal attribute or missing required attribute
    """
    for attribute in attributes:
        required = attributes[attribute][0]
        # Test for required attributes
        if required and attribute not in node.attributes:
            msg = (
                f'"{attribute}" is a required attribute of node "{node.name}"'
            )
            raise MetapypeRuleError(msg)
    for attribute in node.attributes:
        # Test for non-allowed attribute
        if attribute not in attributes:
            msg = f'"{attribute}" is not a recognized attributes of node "{node.name}"'
            raise MetapypeRuleError(msg)
        else:
            # Test for enumerated list of allowed values
            if (
                len(attributes[attribute]) > 1
                and node.attributes[attribute] not in attributes[attribute][1:]
            ):
                msg = f'Node "{node.name}" attribute "{attribute}" must be one of the following: "{attributes[attribute][1:]}"'
                raise MetapypeRuleError(msg)


def validate_children(children: list, node: Node) -> None:
    """
    Validates node children for rule compliance.

    Iterates through the list children rules and validates whether
    the node instance complies with the rules.

    Args:
        children: list of lists containing children
        node: Node instance to be validated

    Returns:
        None

    Raises:
        MetapypeRuleError: Illegal child, bad sequence or choice, missing
        child, or wrong child cardinality
    """
    i = 0
    max_i = len(node.children)
    for child in children:
        name = child[:-2]
        min = child[-2]
        max = child[-1]
        cnt = 0
        while i < max_i:
            child_name = node.children[i].name
            if child_name in name:
                cnt += 1
                if max is not INFINITY and cnt > max:
                    msg = f'Maximum occurrence of "{name}" exceeded for "{node.name}"'
                    raise MetapypeRuleError(msg)
                i += 1
            else:
                break
        if cnt < min:
            msg = f'Minimum occurrence of "{name}" not met for "{node.name}"'
            raise MetapypeRuleError(msg)
    if i < max_i:
        child_name = node.children[i].name
        msg = f'Child "{child_name}" not allowed  for "{node.name}"'
        raise MetapypeRuleError(msg)


dispatcher = {
    names.ACCESS: Accessrule(),
    names.ADDITIONALMETADATA: Additionalmetadatarule(),
    names.ALLOW: Allowrule(),
    names.CONTACT: ResponsiblePartyRule(),
    names.CREATOR: ResponsiblePartyRule(),
    names.DATASET: DatasetRule(),
    names.DENY: DenyRule(),
    names.EML: EmlRule(),
    names.GIVENNAME: Anynamerule(),
    names.INDIVIDUALNAME: IndividualNameRule(),
    names.METADATA: MetadataRule(),
    names.ORGANIZATIONNAME: Anynamerule(),
    names.PERMISSION: PermissionRule(),
    names.POSITIONNAME: Anynamerule(),
    names.PRINCIPAL: PrincipalRule(),
    names.SALUTATION: Anynamerule(),
    names.SURNAME: Anynamerule(),
    names.TITLE: Anynamerule(),
    names.VALUE: ValueRule(),
}
